# flake8: noqa: F401
import importlib
from typing import Iterable

from . import data, functools, path
from .cache_utils import *
from .devices import *
from .dtype import parse_dtype
from .instantiate import instantiate
from .misc import *
from .packages import import_object
from .parameters import *
from .timer import timeit_context



"""
This module contains the implementation of ExPO merge for general nn.Modules.

Reference:
- Zheng et al. Weak-to-Strong Extrapolation Expedites Alignment.
"""

import logging

from fusion_bench import BaseAlgorithm, BaseModelPool
from fusion_bench.method import SimpleAverageAlgorithm
from fusion_bench.utils.state_dict_arithmetic import (
    state_dict_add,
    state_dict_mul,
    state_dict_sub,
)

log = logging.getLogger(__name__)


class ExPOAlgorithm(BaseAlgorithm):
    R"""
    ExPO merge algorithm.

    This algorithm merges a pretrained model with a finetuned model.

    $$\theta_{merged} = \theta_{sft} + \alpha (\theta_{rlhf} - \theta_{sft})$$

    where $\theta_{merged}$ is the merged model, $\theta_{rlhf}$ is the finetuned model (medium-aligned model),
    $\theta_{sft}$ is the pretrained model (base model), and $\alpha$ is the extrapolation factor.

    In the configuration, the SFT model should have name `_pretrained_` and the rlhf name can be set arbitarily.
    """

    _config_mapping = BaseAlgorithm._config_mapping | {
        "extrapolation_factor": "extrapolation_factor"
    }

    def __init__(self, extrapolation_factor: float, **kwargs):
        self.extrapolation_factor = extrapolation_factor
        super().__init__(**kwargs)

    def run(self, modelpool: BaseModelPool):
        """
        Run the ExPO merge algorithm.

        Args:
            modelpool (BaseModelPool): The pool of models to merge.

        Returns:
            nn.Module: The merged model.
        """
        if not isinstance(modelpool, BaseModelPool):
            modelpool = BaseModelPool(modelpool)

        assert len(modelpool.model_names) >= 1, "ExPO requires at least one model."
        assert modelpool.has_pretrained, "ExPO requires pretrained models (base model)."

        sft_model = modelpool.load_pretrained_model()
        if len(modelpool) == 1:
            rlhf_model = modelpool.load_model(modelpool.model_names[0])
        else:
            # if there are multiple RLHF models, use simple average to merge them before running ExPO
            log.info(
                f"There are {len(modelpool)} models in the model pool, averaging them first..."
            )
            rlhf_model = SimpleAverageAlgorithm().run(modelpool)

        # merge the pretrained model and the finetuned model
        delta_parameters = state_dict_sub(
            rlhf_model.state_dict(), sft_model.state_dict()
        )
        merged_sd = state_dict_add(
            rlhf_model.state_dict(),
            state_dict_mul(delta_parameters, scalar=self.extrapolation_factor),
        )

        rlhf_model.load_state_dict(merged_sd)
        return rlhf_model



from typing import Dict, List, Optional

import torch
import torch.nn.functional as F
from torch.nn.utils.rnn import pad_sequence


def padded_collate_sft(
    batch: List[Dict[str, List[int]]],
    padding_idx: int = 0,
    input_ids_key: str = "input_ids",
    attention_mask_key: Optional[str] = "attention_mask",
    labels_key: Optional[str] = "labels",
    ignore_idx: int = -100,
) -> Dict[str, torch.Tensor]:
    """
    Pad (right) a batch of sequences to the longest sequence length in the batch, and
    convert integer lists to tensors.

    Args:
        batch (List[Dict[str, List[int]]]): A list of dictionaries containing input, label pairs.
        padding_idx (int): Padding index for input ids. Defaults to 0.
        ignore_idx (int): Padding index for labels. Defaults to -100.

    Returns:
        Dict[str, torch.Tensor]: Collated input and label tensors.
    """
    input_ids = pad_sequence(
        [torch.tensor(x[input_ids_key]) for x in batch],
        batch_first=True,
        padding_value=padding_idx,
    )
    if attention_mask_key is not None and attention_mask_key in batch[0]:
        attention_mask = pad_sequence(
            [torch.tensor(x[attention_mask_key]) for x in batch],
            batch_first=True,
            padding_value=0,
        )
    else:
        attention_mask = None
    labels = pad_sequence(
        [torch.tensor(x[labels_key]) for x in batch],
        batch_first=True,
        padding_value=ignore_idx,
    )

    if attention_mask is not None:
        collated_batch = {
            input_ids_key: input_ids,
            attention_mask_key: attention_mask,
            labels_key: labels,
        }
    else:
        collated_batch = {input_ids_key: input_ids, labels_key: labels}

    for key in batch[0]:
        if key not in [input_ids_key, attention_mask_key, labels_key]:
            collated_batch[key] = [x[key] for x in batch]

    return collated_batch



# flake8: noqa F401
import sys
from typing import Dict, List, Mapping, TypeVar

from typing_extensions import TypeAlias

try:
    import torch
    from torch import Tensor

    StateDictType: TypeAlias = Dict[str, Tensor]
except ImportError:
    pass


ModuleType = type(sys)
T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")

__all__ = ["StateDictType", "ModuleType", "T", "T1", "T2", "T3", "T4"]
